

# Initialize  -------------------------------------------------------------
library(data.table)
library(stats)
library(plyr)
library(forecast)
library(ggplot2)
library(reshape2)
library(NISTunits)
library(vars)
library(rdist)
library(VGAM)

# Read Data ---------------------------------------------------------------

f_read_data<-function(measurement="TG", years=seq(1,6), var="speed", agg=6){
	# reading the data: We used this function to aggregate the original (10-min)
	# data into hourly data
	dt_all = list()
	for (i in years){
		tmp = fread(paste(data_dir,var,sprintf("_year%i",i),".csv", sep=""))
		n = nrow(tmp)
		tmp = setDT(tmp)[, as.list(colMeans(.SD, na.rm=TRUE)), by = gl(ceiling(n/agg), agg, n)]
		tmp[,gl:=NULL]
		var_name = paste("year",i,sep="")
		dt_all[[var_name]] = tmp
		print(sprintf("read year %i for var %s-- aggregation = %i", i, var, agg))
	}
	dt_all
}

f_latlng_tocar <- function(lat, lon, r = 6371) {

	# This function changes latitude longitude to cartesian coordinate
	# r = 6371; %  km radius of earth

	lat = NISTdegTOradian(lat)
	lon = NISTdegTOradian(lon)
	x = r * cos(lat) * cos (lon)
	y = r * cos(lat) * sin(lon)
	c(x, y)

}

f_calc_masked_coor <- function(TG_coor) {
	# To maintain confidentiality, we use classical multidimensional scaling
	# on the original lat-lng data.  
	n_turb = nrow(TG_coor)
	coor = matrix(0, nrow = n_turb, ncol = 2)
	for (i in 1:n_turb){
		car = f_latlng_tocar(TG_coor[i, lat], TG_coor[i, lon])
		coor[i,] = car
	}
	Dcoor = matrix(0, nrow = n_turb, ncol = n_turb)
	for (i in 1:n_turb){
		for (j in i:n_turb){
			Dcoor[i,j] = sqrt(sum((coor[i,] - coor[j,])^2))
		}
	}
	Dcoor = Dcoor + t(Dcoor)
	cmdscale(Dcoor)
}



calc_rho<-function(dt, beta=0.85, differencing = TRUE){
	# finding corrolation and informative neighborhoos
	col_max = apply(dt, 2, function(x) max(x, rm.na = TRUE))
	col_names = colnames(dt)
	T = nrow(dt)
	z = data.table(t(t(dt) / col_max))
	if (differencing){
		print("taking differences...")
		
		for(i in col_names) {
			z[, s := shift(get(i))]
			z[,(i) := get(i)-s]
		}
		z[,s:=NULL]
	}
	z_bar = colMeans(z,na.rm=TRUE)
	z_dmean = data.table(t(t(z) - z_bar))
	

	z_std = apply(z, 2, sd, na.rm=TRUE)
	z_dmean = data.table(t(t(z_dmean) / z_std)) 

	z_dmean[is.na(z_dmean)] <- 0
	rho = (t(as.matrix(z_dmean))%*%as.matrix(z_dmean))/T
	col_max_rho = apply(rho, 2, function(x) max(x))
	rho = data.table(t(t(rho) / col_max_rho))

	rho[rho<beta] = 0
	rho = as.matrix(rho)
	rho

}


get_TG_coor<-function(filename="TG_coord.csv"){
	TG_coor=fread(paste(data_dir,filename,sep=""))
	TG_coor
}

calc_A<-function(TG_coor, turbine = 1, lambda = c(10,20), p=1){
	n_turb = nrow(TG_coor)
	A=matrix(nrow=n_turb, ncol=p)

	target_coor = TG_coor[TG_coor[,3]==turbine] 
	for (i in 1:n_turb){
		tmp = TG_coor[i,1:2]-target_coor[1:2]
		tmp = tmp^2*lambda
		tmp = sum(tmp)
		tmp = exp(-tmp)
		A[i,] = tmp
	}
	A
}

calc_D<-function(lambda=2, p=1){
	d = seq(1:p)
	d = exp(-d*lambda)
	if (p>1){
		d=diag(d)
	}	
	d
}

create_Y<-function(dt, t, p=1){
	I = ncol(dt)
	
	Y = t(as.matrix(dt[(t-p+1):t-1,]))
	Y = t(apply(t(Y), 2, rev))
	
	
	if (p==1){
		Y = t(Y)
	}
	Y
}

calc_pen<-function(A, rho, beta=0.85){
	M= 1/beta
	rho_inv = 1/rho
	rho_inv[!is.finite(rho_inv)] = M
	pen = t(A)%*%rho_inv
	pen = sum(diag(t(pen)%*%pen))
	pen
}




mean_model<-function(A, D, Y, c = 0, truncate=TRUE, I_tr = 10){
	if (truncate){
		Y = as.matrix(Y[rank(-A)<I_tr], ncol=1)
		A = as.matrix(A[rank(-A)<I_tr], ncol=1)
	}
	AD = A%*%D
	mu = sum(diag(AD%*%t(Y)))
	mu = mu + c
	mu

}


opt_model<-function(lambda, all_Y, yi_e, I, T, TG_coor, rho,
	turbine = 1, h = 2, loss = 'mse', gamma = 0, p = 1, pred_func = 'mean', 
	alpha = 0.73, epoch=0, truncate = TRUE){

	A = calc_A(TG_coor,turbine, lambda[1:2], p)
	D = calc_D(lambda[3], p)
	pen = calc_pen(A, rho)
	mean_for_one_t<-function(all_Y,t){
		Y = as.matrix(all_Y[((t-p-1)*I+1):((t-p)*I),], nrow=I)		
		mean_model(A, D, Y, truncate = truncate)
	}
	mean_pred = sapply(seq(p+1,T-h+1),mean_for_one_t,all_Y = all_Y)
	# err = (yi[(p+h):T] - yibar) - mean_pred
	
	if (pred_func=="mean"){
		yhat = mean_pred
	}

	dt_error = data.table(y = yi_e[(p+h):T,1,with=FALSE], yhat = yhat, epoch=yi_e[(p+h):T,2,with=FALSE])
	names(dt_error)=c("y","yhat", "epoch")

	if (loss=='mse'){
		if(epoch!=0){
			e = epoch
			dt_error = dt_error[epoch==e]
		}
		err = (dt_error$y - dt_error$yhat)	
		err = sum(err^2)/(T-h)
	}
	else{
		if(epoch!=0){
			e = epoch
			dt_error = dt_error[epoch==e]
		}			 
		dt_error[,pce :=apply(dt_error, 1, function(x) pc_error(x[1], x[2], alpha = alpha) )]	
		err = mean(dt_error$pce)
	}
	
	err = err + gamma * pen/10000
	err

}



f_pred_mean<-function(dt, tpred, TG_coor, lambda=c(0.1,0.1,0.8), turbine = 1, h = 2, p = 1){
	# '''
	# tpred: the index of time for prediction, e.g. if tpred is 710 you should compare pred_mean with value at 710
	# '''
	Y = create_Y(dt, tpred-h+1)
	A = calc_A(TG_coor,turbine, lambda[1:2], p)

	D = calc_D(lambda[3], p)
	pred_mean = mean_model(A, D, Y)
	pred_mean

}
	




power_curve<-function(speed, cutin = 3.5, cutrated = 13,  cutout = 20, form = 'linear'){
	if (form !='linear'){
		stop("to be added")
	}
	if ((speed < cutin) || (speed > cutout)) {
		power = 0
	} else if (speed <= cutrated){
		power = (1/(cutrated - cutin)*(speed - cutin))
	} else {
		power = 1
	}
	power

}


f_crps <- function(stest, smean, sstd, gi = 0.5, cutin = 3.5, cutrated = 13, alpha = 0.73) {


	f_phi <- function(y) { 0.5 * (1 + erf(y / sqrt(2)))}

	F_s <- function(s) {
		num_ <- f_phi( (s - smean)/sstd) - f_phi(-smean / sstd)
		den_ <- 1  - f_phi(-smean / sstd)
		return(num_/den_)
	}

	ginv <- function(x) {
		out <- (cutrated-cutin) * x + cutin

		if (x>=1) {
			return(cutrated)
		}
		if (x<=0) {
			return(cutin)
		}
		else {
			return(out)
		}
	}

	# si = ginv(gi)

	integrand1 <- function(x) {
		F_s(ginv(x))^2 
	}

	integrand2 <- function(x) {
		(F_s(ginv(x))-1)^2 
	}

	gi = power_curve(stest)
	a1 = alpha * integrate(Vectorize(integrand1),0 , gi)[[1]]
	a2 = (1-alpha) * integrate(Vectorize(integrand2),gi , 1)[[1]]

	return(2*(a1+a2))
}

pc_error<-function(speed, pred, alpha = 0.73, cutout = 20){
	if ((pred<= speed) & (speed <= cutout)){
		pce = alpha*(power_curve(speed)-power_curve(pred))
	} else if ((pred> speed) & (pred <= cutout)) {
		pce = (1-alpha)*(power_curve(pred)-power_curve(speed))
	} else if ((speed > cutout) & (pred > cutout)) {
		pce = 0
	} else if ((speed > cutout) & (pred < cutout)){
		pce = alpha
	} else {
		pce = 1 - alpha
	}

	pce
}


find_neighborhood<-function(rho,turbine){
	x = data.table(which(rho>0, arr.ind=TRUE))
	J = x[col==turbine,.(row)]
	J
}




calc_nu <-function(dt, rho, start = 1, end = T, turbine = 1){
	print(sprintf("calculating nu for turbine %i",turbine))
	J = find_neighborhood(rho,turbine)

	nu_for_one_t<-function(dt,J, t){

		y = dt[,paste("V",J$row,sep=""),with=FALSE]

		tryCatch(
		{
			tmp = sum((y[t,]-y[t-1,])^2+(y[t-1,]-y[t-2,])^2)
			nu_t = sqrt(tmp/(2*nrow(J)))
			nu_t
		},
		error = function(cond)
		{
			message(cond)
			print("data missing...returning NA")
			NA
			})#end tryCatch
	}


	nu = sapply(seq(start,end), nu_for_one_t, dt = dt, J = J)
	nu
}

var_model<-function(dt, std_e, rho, start = 1, end = T, turbine = 1, epoch = 0){

	nu = calc_nu(dt, rho, start = start, end = end, turbine = turbine)

	if (epoch != 0){
		e = epoch
		nu = nu[which(std_e$epoch==e)]
		std = std_e[epoch ==e, 1]
	}
	else{
		std = std_e[,1]
	}
	co_std = lsfit(nu, std)
	B = co_std$coefficients
	B
}

f_quan_pred <- function(m, s, alpha = 0.73){
	return (m+alpha*qnorm(alpha + (1-alpha)*pnorm(-m/s)))
}

f_do_optimization <- function(dt_month_epoch, T, TG_coor, rho, turbine = 1, epoch = 1, h = 2, initial_values = c(0.1,0.2,0.6),
 p = 1, alpha = 0.73, gamma = 0){
	
	I = nrow(TG_coor)
	print(sprintf(" turbine=%i, epoch= %i, h=%i, p=%i, alpha=%f, I=%i", 
		turbine, epoch, h, p, alpha, I))
	yi_e = dt_month_epoch[1:T,c(paste("V",turbine,sep=""),"epoch"), with=FALSE]
	all_Y = ldply(seq(p+1,T),create_Y,dt=dt_month_epoch[,-ncol(dt_month_epoch),with=FALSE])


	o = optim(par = c(0.1,0.2,0.6),opt_model, all_Y = all_Y, yi_e = yi_e, I = I,  T = T, 
	TG_coor = TG_coor, rho = rho, turbine = turbine, h =h, loss='pc_error',
	gamma = gamma, p = p, alpha = alpha, epoch = epoch, 
	method="L-BFGS-B",lower=c(0.0001,0.0001,0.002), upper=c(3,3,3),control=list(trace=1, REPORT=1))
	o

}	


f_estimate_std <- function (std_month_epoch, dt_month_epoch, T, turbine = 1, epoch = 1, 
	test_period = 24){
	print(sprintf(" turbine=%i, epoch= %i", turbine, epoch))

	B = var_model(dt_month_epoch[1:T,-ncol(dt_month_epoch),with=FALSE], 
		std_month_epoch[1:T, c(paste("V",turbine,sep=""),"epoch"), with=FALSE], rho,
		start = 1, end = T, 
		turbine = turbine, epoch = epoch)

	nu_p = calc_nu(dt_month_epoch[,-ncol(dt_month_epoch),with=FALSE], rho, 
		start = T+1, end = T+test_period, turbine = turbine)
	std_pred = cbind(rep(1,test_period), nu_p)%*%as.matrix(B, ncol=1,nrow=2)
	std_pred
}

f_calc_error <- function(dt_month_epoch, o, std_pred, T, TG_coor, rho, 
	turbine = 1, epoch = 1, h =2, 
	test_period = 24, p = 1, alpha = 0.73, cutout = 20, month = 30*24, calc_pers = TRUE) {
	
	I = nrow(TG_coor)
	print(sprintf(" turbine=%i, epoch= %i, test_period=%i, alpha=%f, I=%i", 
		turbine, epoch, test_period, alpha, I))
	dt_test = dt_month_epoch[(T+1):(T+test_period),c(paste("V",turbine,sep=""),"epoch"), with=FALSE]

	
	pred = rep(0, test_period)

	c = 0
	for (tpred in seq(T+1,T+test_period)){
		c = c +1
		pred[c] = f_pred_mean(dt_month_epoch[,-ncol(dt_month_epoch),with=FALSE], tpred, 
			TG_coor, lambda=o$par, turbine = turbine, h =h, p = p)

	}

	print("calculating gstar error...")
	pred = mapply(f_quan_pred, pred, std_pred)

	# if prediction is higher than the cutout speed, we just use the cutout speed
	pred[pred>cutout] <- cutout

	dt_pred = data.table(t=seq(T+1,T+test_period), y = as.numeric(as.matrix((dt_test[,1]))), 
		epoch=dt_test$epoch, pred=pred)

	# sum(pred-dt_test)^2/test_period
	if (epoch!=0){
		e = epoch
		dt_pred = dt_pred[epoch==e]
	}
	
			

	# persistence error
	if (calc_pers){
		print("calculating persistence error...")
		cols_per_pred = paste("per", colnames(dt_month_epoch), sep="_")

		dt_pers = copy(dt_month_epoch)
		dt_pers[, (cols_per_pred) := shift(.SD, h, NA, "lag"), .SDcols=colnames(dt_pers)]
		pred_per = dt_pers[seq(T+1,T+test_period),c(paste("per_V",turbine,sep=""),"epoch"), with=FALSE]

		if (epoch!=0){
			e = epoch
			pred_per = pred_per[epoch==e]
		}
		
		pred_per_num = as.numeric(as.matrix(pred_per[,paste("per_V",turbine,sep=""),with=FALSE]))
	} else{
		pred_per_num = 0
	}
	
	
	dt_error = data.table(y = dt_pred$y, yhat = dt_pred$pred, yper = pred_per_num)
	
	dt_error[,gstar_error :=apply(dt_error, 1, function(x) pc_error(x[1], x[2], alpha = alpha) )]
	if (calc_pers){dt_error[,pers_error :=apply(dt_error, 1, function(x) pc_error(x[1], x[3], alpha = alpha) )]}
	
	print(sprintf("error for turbine=%i, gstar=%f------persistence=%f",turbine, mean(dt_error$gstar), mean(dt_error$pers)))
	list(dt_error=dt_error, dt_pred = dt_pred)

}

f_calc_arma_error <- function(dt_month_epoch, T, turbine = 1, epoch = 1,
	h = 2, test_period = 24, order = c(1, 0, 2), do_seasonal_training = TRUE, alpha = 0.73, crps = 0) {

	print(sprintf("ARMA: turbine=%i, epoch= %i, test_period=%i, alpha=%f", 
		turbine, epoch, test_period, alpha))

	dt_test = dt_month_epoch[(T+1):(T+test_period),c(paste("V",turbine,sep=""),"epoch"), with=FALSE]
	if (crps==1) {do_seasonal_training = FALSE} #for faster training when calculating crps; change this to TRUE for progressive training
	pred_arma = rep(0, test_period)
	

	c = 0
	print("ARMA")
	for (tpred in seq(T+1,T+test_period)){
		c = c +1
		ts_obj = ts(dt_month_epoch[1:(tpred-h),1])
		arima_model = Arima(ts_obj, order = order)
		arima_pred = predict(arima_model, h)
		pred_arma[c] = arima_pred[[1]][h]
		

	}
	c = 0 
	pred_seas = rep(0, test_period)
	
	if (do_seasonal_training){
		print("Seasonal ARMA: train progressively...")
		print("(Use do_seasonal_training = FALSE for one-time training)")
		for (tpred in seq(T+1,T+test_period)){
			c <- c + 1
			ts_obj = ts(dt_month_epoch[1:(tpred-h),1], freq = 24)
			arima_model_seas = Arima(ts_obj, order = order, seasonal=list(order=order, period=24))
            
			arima_pred_seas = predict(arima_model_seas, h)
			pred_seas[c] = arima_pred_seas[[1]][h]
			if (c%%6==0){
				print(sprintf("   c=%i",c))
			}
		}	
	} else {
		print("Seasonal ARMA: train once")
		ts_obj = ts(dt_month_epoch[1:T,1], freq = 24)
		arima_model_seas = Arima(ts_obj, order = order, seasonal=list(order=order, period=24))
		arima_pred_seas = predict(arima_model_seas, test_period)
		pred_seas = arima_pred_seas[[1]][1:test_period]
		std_seas = arima_pred_seas[[2]][1:test_period]
	}
	print("calculating ARMA error...")
	dt_test = dt_month_epoch[(T+1):(T+test_period),c(paste("V",turbine,sep=""),"epoch"), with=FALSE]
	dt_pred = data.table(t=seq(T+1,T+test_period), y = as.numeric(as.matrix((dt_test[,1]))), 
		epoch=dt_test$epoch, pred_arma=pred_arma, pred_arma_seas=pred_seas)

	if (crps ==1 ){
		dt_pred = data.table(t=seq(T+1,T+test_period), y = as.numeric(as.matrix((dt_test[,1]))), 
			epoch=dt_test$epoch, pred_arma=pred_arma, pred_arma_seas=pred_seas, std_arma_seas = std_seas)	
	}
	
	
	if (epoch!=0){
		e = epoch
		dt_pred = dt_pred[epoch==e]
	}
	dt_error = data.table(y = dt_pred$y, yarma = dt_pred$pred_arma, yarmaseas = dt_pred$pred_arma_sea)
	
	dt_error[,arma_error :=apply(dt_error, 1, function(x) pc_error(x[1], x[2], alpha = alpha) )]
	dt_error[,arma_seas_error :=apply(dt_error, 1, function(x) pc_error(x[1], x[3], alpha = alpha) )]
	
	print(sprintf("error for turbine=%i, arma=%f------arma_seasonal=%f",turbine, 
		mean(dt_error$arma_error), mean(dt_error$arma_seas_error)))
	list(dt_error=dt_error, dt_pred = dt_pred)

	
}


f_calc_var_error <- function(dt_month_epoch, T, pairwise_dist, turbine = 1, epoch = 1,
	h = 2, test_period = 24, alpha = 0.73, max_dist = 2) {

	# note that the max_dist (=3) here is based on transformed locations
	print(sprintf("VAR: turbine=%i, epoch= %i, test_period=%i, alpha=%f", 
		turbine, epoch, test_period, alpha))

	dt_test = dt_month_epoch[(T+1):(T+test_period),c(paste("V",turbine,sep=""),"epoch"), with=FALSE]

	pred_var = rep(0, test_period)
	neigh <- which(pairwise_dist[turbine,] < max_dist)
	dt <- dt_month_epoch[,c(paste("V",neigh,sep=""), "epoch"),with=FALSE]

	c <- 0
	vturbine <- paste0('V', turbine)
	print("VAR")
	for (tpred in seq(T+1,T+test_period)){
		c = c +1
		ts_obj = ts(dt[1:(tpred-h),-ncol(dt), with =FALSE])
		fitvar1 = VAR(ts_obj, p=1, type="both")

		pred_var[c] <- predict(fitvar1, n.ahead=h)$fcst[[vturbine]][h,1]


	}
	# If VAR produces NA forecasts, replace them with zero.
	pred_var[is.na(pred_var)]<-0

	dt_pred = data.table(t=seq(T+1,T+test_period), y = as.numeric(as.matrix((dt_test[,1]))), 
		epoch=dt_test$epoch, pred_var=pred_var)

	
	if (epoch!=0){
		e = epoch
		dt_pred = dt_pred[epoch==e]
	}
	dt_error = data.table(y = dt_pred$y, yvar = dt_pred$pred_var)
	
	dt_error[,var_error :=apply(dt_error, 1, function(x) pc_error(x[1], x[2], alpha = alpha) )]
	
	
	print(sprintf("error for turbine=%i, var=%f",turbine, 
		mean(dt_error$var_error)))
	list(dt_error=dt_error, dt_pred = dt_pred)
}


